
#define _GNU_SOURCE             /* for asprintf */
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <beryl.h>
#include <glib.h>
#include <dlfcn.h>
#include <errno.h>
#define MIN_DUMP_VER 1

char *backgroundImage = "background.png";

Window currentRoot = 0;
Bool restartSignal = FALSE;

CompWindow *lastFoundWindow = 0;
CompWindow *lastDamagedWindow = 0;

Bool replaceCurrentWm = TRUE;
Bool indirectRendering = FALSE;
Bool strictBinding = FALSE;
Bool useCow = FALSE;


char * programName;
GLushort defaultColor[4] = { 0xffff, 0xffff, 0xffff, 0xffff };
char *windowTypeString[] = {
    N_("Desktop"),
    N_("Dock"),
    N_("Toolbar"),
    N_("Menu"),
    N_("Utility"),
    N_("Splash"),
    N_("Dialog"),
    N_("Normal"),
    N_("DropdownMenu"),
    N_("PopupMenu"),
    N_("Tooltip"),
    N_("Notification"),
    N_("Combo"),
    N_("Dnd"),
    N_("ModalDialog"),
    N_("Fullscreen"),
    N_("Unknown")
};
int nWindowTypeString =
    sizeof (windowTypeString) / sizeof (windowTypeString[0]);

REGION emptyRegion;
REGION infiniteRegion;
int programArgc=0;
char ** programArgv=NULL;
Display * xDisp;
static char *
setTypeToString (CompOptionType type)
{
    switch (type)
    {
    case CompOptionTypeBool:
        return "bool";
    case CompOptionTypeColor:
        return "color";
    case CompOptionTypeInt:
        return "int";
    case CompOptionTypeFloat:
        return "float";
    case CompOptionTypeString:
        return "string";
    case CompOptionTypeList:
        return "list";
    case CompOptionTypeAction:
        return "action";
    default:
        break;
    }

    return "unknown";
}

static char *
getShortDesc (CompOption * o)
{
    return g_strdup (o->shortDesc);
}

static char *
getLongDesc (CompOption * o, const char *option_name)
{
    char *base;

    return g_strdup (o->longDesc);
}

GKeyFile * schemaFile;
static void set_save_option(GKeyFile * schemaFile, gchar * sect, gchar * dval, CompOption * o)
{
    switch(o->type)
    {
        case CompOptionTypeBool:
            g_key_file_set_boolean(schemaFile,sect,dval,o->value.b);
            break;
        case CompOptionTypeInt:
            g_key_file_set_integer(schemaFile,sect,dval,o->value.i);
            break;
        case CompOptionTypeString:
            g_key_file_set_string(schemaFile,sect,dval,o->value.s);
            break;
        case CompOptionTypeFloat:
            {
                gchar s[G_ASCII_DTOSTR_BUF_SIZE];
                g_ascii_dtostr(s,G_ASCII_DTOSTR_BUF_SIZE,o->value.f);
                g_key_file_set_string(schemaFile,sect,dval,s);
            }
            break;
        case CompOptionTypeColor:
            {
                gint c[4];
                c[0]=o->value.c[0];
                c[1]=o->value.c[1];
                c[2]=o->value.c[2];
                c[3]=o->value.c[3];
                g_key_file_set_integer_list(schemaFile,sect,dval,c,4);
            }
            break;
        case CompOptionTypeAction:
            {
                //format...
                //string list
                //button;key;edge;bell;
                //button-format
                //  modifiers(int);button(int);
                //key-format
                //  modifiers(int);keysym(int);
                //edge-format
                //  edgemask(int)
                //bell-format
                //  on_bell(int)
                gint d[6];
                d[0]=o->value.action.button.modifiers;
                d[1]=(o->value.action.type & CompBindingTypeButton)?
                    o->value.action.button.button:0;
                d[2]=o->value.action.key.modifiers;
                d[3]=(o->value.action.type & CompBindingTypeKey)?
                    XKeycodeToKeysym(xDisp,o->value.action.key.keycode,0)
                    :0;
                d[4]=o->value.action.edgeMask;
                d[5]=o->value.action.bell?1:0;
                g_key_file_set_integer_list(schemaFile,sect,dval,d,6);
            }
            break;
        case CompOptionTypeList:
            {
                gint i;
                switch(o->value.list.type)
                {
                    case CompOptionTypeBool:
                        {
                            gboolean * d = malloc(sizeof(gboolean)*o->value.list.nValue);
                            for (i=0;i<o->value.list.nValue;i++)
                                d[i]=o->value.list.value[i].b;
                            g_key_file_set_boolean_list(schemaFile,sect,dval,d,o->value.list.nValue);
                            free(d);
                        }
                        break;
                    case CompOptionTypeInt:
                        {
                            gint * d = malloc(sizeof(gint)*o->value.list.nValue);
                            for (i=0;i<o->value.list.nValue;i++)
                                d[i]=o->value.list.value[i].i;
                            g_key_file_set_integer_list(schemaFile,sect,dval,d,o->value.list.nValue);
                            free(d);
                        }
                        break;
                    case CompOptionTypeString:
                        {
                            gchar ** d = malloc(sizeof(gchar *)*o->value.list.nValue);
                            for(i=0;i<o->value.list.nValue;i++)
                                d[i]=o->value.list.value[i].s;
                            g_key_file_set_string_list(schemaFile,sect,dval,
                                    (const gchar **)d,o->value.list.nValue);
                            free(d);
                        }
                        break;
                    case CompOptionTypeFloat:
                        {
                            gchar ** d = malloc(sizeof(gchar *)*o->value.list.nValue);
                            for (i=0;i<o->value.list.nValue;i++)
                            {
                                d[i]=malloc(G_ASCII_DTOSTR_BUF_SIZE);
                                g_ascii_dtostr(d[i],G_ASCII_DTOSTR_BUF_SIZE,
                                        o->value.list.value[i].f);
                            }
                            g_key_file_set_string_list(schemaFile,sect,dval,
                                    (const gchar **) d,o->value.list.nValue);
                            for (i=0;i<o->value.list.nValue;i++)
                                free(d[i]);
                            free(d);
                        }
                        break;
                    case CompOptionTypeAction:
                        {
                            gint * d = malloc(sizeof(gint)*o->value.list.nValue*6);
                            for(i=0;i<o->value.list.nValue;i++)
                            {

                                d[i*6]=o->value.list.value[i].action.button.modifiers;
                                d[i*6+1]=(o->value.list.value[i].action.type & CompBindingTypeButton)
                                    ?o->value.list.value[i].action.button.button:0;
                                d[i*6+2]=o->value.list.value[i].action.key.modifiers;
                                d[i*6+3]=(o->value.list.value[i].action.type & CompBindingTypeKey)
                                    ?XKeycodeToKeysym(xDisp,
                                            o->value.list.value[i].action.key.keycode,0):0;
                                d[i*6+4]=o->value.list.value[i].action.edgeMask;
                                d[i*6+5]=o->value.list.value[i].action.bell?1:0;
                            }
                            g_key_file_set_integer_list(schemaFile,sect,dval,d,
                                    o->value.list.nValue*6);
                            free(d);
                        }
                        break;
                    case CompOptionTypeColor:
                        {
                            gint * d = malloc(sizeof(gint)*o->value.list.nValue*4);
                            for(i=0;i<o->value.list.nValue;i++)
                            {
                                gint j;
                                for(j=0;j<4;j++)
                                {
                                    d[i*4+j]=o->value.list.value[i].c[j];
                                }
                            }
                            g_key_file_set_integer_list(schemaFile,sect,dval,d,
                                    o->value.list.nValue*4);
                            free(d);
                        }
                        break;
                    case CompOptionTypeList:
                        g_error(_("List type is not a valid list type"));
                }
            }
    }
}

static void
setDumpToSchema (CompOption * o, char *name, char *plugin, gboolean screen, gboolean special)
{
    gchar *sect;
    gchar *basename;
    gchar *cshort, *clong;

    if (plugin && !special)
    {
        sect = plugin;
        if (!screen)
            basename = g_strdup_printf ("a_%s", name);
        else
            basename = g_strdup_printf ("s_%s", name);
    }
    else if (special)
    {
        sect = plugin;
        basename = g_strdup_printf ("p_%s", name);
    }
    else
    {
        sect = "_";
        if (!screen)
            basename = g_strdup_printf ("a_%s", name);
        else
            basename = g_strdup_printf ("s_%s", name);
    }


    g_key_file_set_string (schemaFile, sect,
                           g_strconcat ("o_", basename, NULL),
                           setTypeToString (o->type));
    if (o->type == CompOptionTypeList)
    {
        g_key_file_set_string (schemaFile, sect,
                               g_strconcat ("t_", basename, NULL),
                               setTypeToString (o->value.list.type));
    }
    if (o->type == CompOptionTypeAction)
    {
        gboolean states[] = { FALSE, FALSE, FALSE, FALSE };
        //check for action types here
        if (o->value.action.state & CompActionStateInitButton)
            states[0] = TRUE;
        if (o->value.action.state & CompActionStateInitKey)
            states[1] = TRUE;
        if (o->value.action.state & CompActionStateInitEdge)
            states[2] = TRUE;
        if (o->value.action.state & CompActionStateInitBell)
            states[3] = TRUE;
        g_key_file_set_boolean_list (schemaFile, sect,
                                     g_strconcat ("s_", basename, NULL),
                                     states, 4);
    }
    else if (o->type == CompOptionTypeString
             || (o->type == CompOptionTypeList
                 && o->value.list.type == CompOptionTypeString))
    {
        if (o->rest.s.nString)
        {
            g_key_file_set_string_list (schemaFile, sect,
                                        g_strconcat ("s_", basename, NULL),
                                        (const gchar **)o->rest.s.string,
                                        o->rest.s.nString);
        }
    }
    else if (o->type == CompOptionTypeInt)
    {
        gint n[] = { o->rest.i.min, o->rest.i.max };
        g_key_file_set_integer_list (schemaFile, sect,
                                     g_strconcat ("s_", basename, NULL),
                                     n, 2);
    }
    else if (o->type == CompOptionTypeFloat)
    {
        gchar *s[3];
        gint i;
        for (i = 0; i < 3; i++)
        {
            s[i] = malloc (G_ASCII_DTOSTR_BUF_SIZE);
        }
        g_ascii_dtostr (s[0], G_ASCII_DTOSTR_BUF_SIZE, o->rest.f.min);
        g_ascii_dtostr (s[1], G_ASCII_DTOSTR_BUF_SIZE, o->rest.f.max);
        g_ascii_dtostr (s[2], G_ASCII_DTOSTR_BUF_SIZE, o->rest.f.precision);
        g_key_file_set_string_list (schemaFile, sect,
                                    g_strconcat ("s_", basename, NULL),
                                    (const gchar **)s, 3);
    }
    gchar *dval = g_strconcat ("d_", basename, NULL);
    if (strcmp (sect, "_") == 0 && strcmp (basename, "a_active_plugins") == 0)
        dval = g_strconcat ("h_", basename, NULL);
    set_save_option (schemaFile, sect, dval, o);

    cshort = getShortDesc (o);
    clong = getLongDesc (o, name);
    g_key_file_set_string (schemaFile, sect,
                           g_strconcat ("n_", basename, NULL), cshort);
    g_key_file_set_string (schemaFile, sect,
                           g_strconcat ("l_", basename, NULL), clong);
}


static void dumpOptions(CompOption * o, int no, char * plugin, gboolean screen)
{
    int i;
    for (i=0;i<no;i++)
    {
        setDumpToSchema (&o[i], o[i].name,plugin,screen,FALSE);
    }
}
static CompPlugin * setLoadPlugin(char * pname)
{
    CompPlugin * p = malloc(sizeof(CompPlugin));
    void * dlhand;
    int e;
    dlerror();
    dlhand = dlopen (pname, RTLD_LAZY);
    e=errno;
    if (dlhand)
    {
        PluginGetInfoProc getInfo;
        char * error;
        dlerror();
        getInfo = (PluginGetInfoProc) dlsym(dlhand, "getCompPluginInfo");
        error = dlerror();
        if (error)
        {
            puts(error);
            exit(1);
        }
        p->vTable = getInfo();
        p->devPrivate.ptr = dlhand;
        p->devType = "dlloader";
    }
    else
    {
        char * er = dlerror();
        if (er)
            puts(er);
        puts(strerror(e));
        exit(1);
    }
    return p;
}
static void dumpPluginOptions(CompPlugin * p)
{
    CompOption info, bopt, ropt, copt;
    GArray *reqs, *before, *con;

    memset (&info, 0, sizeof (info));
    info.name = "info";
    info.shortDesc = p->vTable->shortDesc;
    info.longDesc = p->vTable->longDesc;
    info.type = CompOptionTypeString;
    info.value.s = "";
    setDumpToSchema (&info, info.name, p->vTable->name, FALSE,TRUE);

    reqs = g_array_new (FALSE, FALSE, sizeof (CompOptionValue));
    before = g_array_new (FALSE, FALSE, sizeof (CompOptionValue));
    con = g_array_new(FALSE,FALSE, sizeof (CompOptionValue));

    if (p->vTable->deps)
    {
        int i;
        CompOptionValue value;

        for (i = 0; i < p->vTable->nDeps; i++)
        {
            value.s = p->vTable->deps[i].plugin;
            switch(p->vTable->deps[i].rule)
            {
                case CompPluginRuleBefore:
                    g_array_append_val (before, value);
                    break;
                case CompPluginRuleConflicts:
                    g_array_append_val (con, value);
                    break;
                case CompPluginRuleAfterAlt:
                case CompPluginRuleAfter:
                    g_array_append_val (reqs, value);
                    break;
            }
        }
    }

    memset (&copt, 0, sizeof (copt));
    copt.name = "conflicts";
    copt.shortDesc = N_("Plugins that this conflicts with");
    copt.longDesc = N_("Do not modify");
    copt.type = CompOptionTypeList;
    copt.value.list.type = CompOptionTypeString;
    copt.value.list.nValue = con->len;
    copt.value.list.value = (CompOptionValue *) con->data;
    setDumpToSchema (&copt, copt.name, p->vTable->name, FALSE,TRUE);

    memset (&bopt, 0, sizeof (bopt));
    bopt.name = "load_before";
    bopt.shortDesc = N_("Plugins that this must load before");
    bopt.longDesc = N_("Do not modify");
    bopt.type = CompOptionTypeList;
    bopt.value.list.type = CompOptionTypeString;
    bopt.value.list.nValue = before->len;
    bopt.value.list.value = (CompOptionValue *) before->data;
    setDumpToSchema (&bopt, bopt.name, p->vTable->name, FALSE,TRUE);

    memset (&ropt, 0, sizeof (ropt));
    ropt.name = "requires";
    ropt.shortDesc = N_("Plugins that this requires");
    ropt.longDesc = N_("Do not modify");
    ropt.type = CompOptionTypeList;
    ropt.value.list.type = CompOptionTypeString;
    ropt.value.list.nValue = reqs->len;
    ropt.value.list.value = (CompOptionValue *) reqs->data;
    setDumpToSchema (&ropt, ropt.name, p->vTable->name, FALSE,TRUE);
}
int main(int argc, char ** argv)
{
    emptyRegion.rects = &emptyRegion.extents;
    emptyRegion.numRects = 0;
    emptyRegion.extents.x1 = 0;
    emptyRegion.extents.y1 = 0;
    emptyRegion.extents.x2 = 0;
    emptyRegion.extents.y2 = 0;
    emptyRegion.size = 0;

    infiniteRegion.rects = &infiniteRegion.extents;
    infiniteRegion.numRects = 1;
    infiniteRegion.extents.x1 = MINSHORT;
    infiniteRegion.extents.y1 = MINSHORT;
    infiniteRegion.extents.x2 = MAXSHORT;
    infiniteRegion.extents.y2 = MAXSHORT;

    if (argc) programName=argv[0];
    int i;
    CompOption * o;
    xDisp = XOpenDisplay(getenv("DISPLAY"));
    if (!xDisp)
        exit(1);
    schemaFile = g_key_file_new();
    if (argc>1)
    {
        int n;
        for (n=1;n<argc;n++)
        {
            CompPlugin * p = setLoadPlugin(argv[n]);
            dumpPluginOptions(p);
            if (p->vTable->version<MIN_DUMP_VER)
                exit(1);
            if (p->vTable->getDisplayOptions)
            {
                o = p->vTable->getDisplayOptions(NULL,&i);
                dumpOptions(o,i,p->vTable->name,FALSE);
            }
            if (p->vTable->getScreenOptions)
            {
                o = p->vTable->getScreenOptions(NULL,&i);
                dumpOptions(o,i,p->vTable->name,TRUE);
            }
        }
    }
    else
    {
        o = compGetDisplayOptions(NULL,&i);
        dumpOptions(o,i,NULL,FALSE);
        o = compGetScreenOptions(NULL,&i);
        dumpOptions(o,i,NULL,TRUE);
    }
    puts(g_key_file_to_data(schemaFile,NULL,NULL));
    return 0;
}
